// =================================================================================================
//
//   File....... dc22_badge_master.cpp
//               -- adapted by David Betz from dc22_badge_master.spin
//               -- for internal use only!
//               -- no changes unless approved/reviewed by JonnyMac/Ryan
// 
//   Authors.... Jon "JonnyMac" McPhalen and Ryan "1o57" Clarke
//               Converted to C++ by David Betz
//               MIT License
//               -- see below for terms of use
//
//   E-mail..... jon@jonmcphalen.com
//               1o57@10000100001.org 
//
// =================================================================================================

/*

  Welcome to Defcon 22. This year we would like to invite you to experiment more fully with your
  badge -- feel free to play around with code.

  You can load directly to RAM [F10] if you don//t want to blast your firmware, but even if you do,
  we are giving you the source from the start. The source provides a nice badge template with extra
  objects so that you can experiment with LEDs, buttons, IR (in and out), timing, speed changes, etc.

  Completing the challenge will at some point require you to //update// your badge -- but for now, how
  about changing your LED pattern? It//s easier than you think! If you need help, feel free to stop
  by the Hardware Hacking Village, or simply ask someone who has a different pattern than yours.
  Create a new pattern -- have fun!

*/

#include <stdint.h>
#include <propeller.h>

#define ABS(x)  ((x) >= 0 ? (x) : -(x))

#if 0

/* timing */

  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000                                          // use 5MHz crystal

  CLK_FREQ = ((_clkmode - xtal1) >> 6) * _xinfreq               // system freq as a constant
  MS_001   = CLK_FREQ / 1_000                                   // ticks in 1ms
  US_001   = CLK_FREQ / 1_000_000                               // ticks in 1us

#endif

// speed settings for power control/reduction
// -- use with clkset() instruction

#define XT1_P16  0x6f // %0_1_1_01_111                          // 16x crystal (5MHz) = 80MHz
#define XT1_PL8  0x6e // %0_1_1_01_110 
#define XT1_PL4  0x6d // %0_1_1_01_101                                      
#define XT1_PL2  0x6c // %0_1_1_01_100
#define XT1_PL1  0x6b // %0_1_1_01_011
#define RC_SLOW  0x01 // %0_0_0_00_001                          // 20kHz


// program speed and terminal baud
  
#define B_SPEED  20     /* MHz */
#define T_BAUD   57600  /* for terminal io */

#define IR_FREQ  36000  /* matches receiver on DC22 badge */
#define IR_BAUD  2400   /* max supported using IR connection */
#define IR_BLAST 0xdc22

enum {
  HUMAN,
  ARTIST,
  PRESS,
  SPEAKER,
  VENDOR,
  CONTEST,
  GOON,
  UBER,
  LO57
};

const int BADGE_TYPE = HUMAN;

/* io pins */

#define RX1    31                                               // programming / terminal
#define TX1    30
  
#define SDA    29                                               // eeprom / i2c
#define SCL    28

#ifdef QUICKSTART
#define PAD3   5                                                // touch pads (QuickStart)
#define PAD2   4
#define PAD1   3
#define PAD0   2
#else
#define PAD3   27                                               // touch pads (DEFCON 22 Badge)
#define PAD2   26
#define PAD1   25
#define PAD0   24
#endif
 
#define LED7   23                                               // leds
#define LED6   22
#define LED5   21
#define LED4   20
#define LED3   19
#define LED2   18
#define LED1   17
#define LED0   16

#define IR_IN  15                                               // ir input
#define IR_OUT 14                                               // ir output

#if 0

con /* io configuration */

  IS_OFF    =  0                                                // all bits off   
  IS_ON     = -1                                                // all bits on    

  IS_LOW    =  0                                                
  IS_HIGH   = -1                                                

  IS_INPUT  =  0
  IS_OUTPUT = -1
  
#endif

/* pst formatting */

enum {
  HOME = 1,
  GOTOXY,
  BKSP = 8,
  TAB,
  LF,
  CLREOL,
  CLRDN,
  CR,
  GOTOX = 14,
  GOTOY,
  CLS
};

#if 0

obj

  ee   : "jm_24xx512"                                           // eeprom access
  pwm  : "jm_pwm8"                                              // pwm for LEDs
 
#endif

#include "CryptoFdSerial.h"
CryptoFdSerial term;                                            // crypto serial io for terminal

#include "jm_SircsTx.h"
jm_SircsTx irtx;                                                // SIRCS output

#include "jm_SircsRx.h"
jm_SircsRx irrx;                                                // SIRCS input

#include "jm_Prng.h"
jm_Prng prng;                                                   // random #s

#include "jm_EzTimer.h"
jm_EzTimer tmr1;                                                // asynchronous timer

#include "jm_Animation.h"
jm_Animation animation(LED0, 8);                                // LED animation

int ms001;                                                      // system ticks per millisecond
int us001;                                                      // system ticks per microsecond

void test_ir();
void alien_badge();
int check_ir();
void setup();
void set_speed(int mhz);
void pause(int ms);
void high(int pin);
void low(int pin);
void toggle(int pin);
int input(int pin);
int read_pads();

// Animation tables for LEDs
// -- 1st byte is number of steps in animation sequence
// -- each step holds pattern and hold time (ms)
// -- for delays > 255, duplicate pattern + delay

uint8_t Cylon[] = {
              14,  // number of rows of data below
              /* %10000000 */ 0x80, 125,
              /* %01000000 */ 0x40, 125,
              /* %00100000 */ 0x20, 125,
              /* %00010000 */ 0x10, 125,
              /* %00001000 */ 0x08, 125, 
              /* %00000100 */ 0x04, 125, 
              /* %00000010 */ 0x02, 125,
              /* %00000001 */ 0x01, 125,
              /* %00000010 */ 0x02, 125,
              /* %00000100 */ 0x04, 125,
              /* %00001000 */ 0x08, 125,  
              /* %00010000 */ 0x10, 125,  
              /* %00100000 */ 0x20, 125,    
              /* %01000000 */ 0x40, 125
};
                    
uint8_t Chaser[] = {
              3,    // number of rows of data below
              /* %10010010 */ 0x92,  75,
              /* %00100100 */ 0x24,  75,
              /* %01001001 */ 0x49,  75
};

uint8_t InOut[] = {
              6,   // number of rows of data below
              /* %10000001 */ 0x81, 100,
              /* %01000010 */ 0x42, 100,
              /* %00100100 */ 0x24, 100,
              /* %00011000 */ 0x18, 100,
              /* %00100100 */ 0x24, 100,
              /* %01000010 */ 0x42, 100
};

uint8_t Police[] = {
              8,   // number of rows of data below
              /* %11001100 */ 0xcc,  75,
              /* %11110000 */ 0xf0,  75,
              /* %11001100 */ 0xcc,  75,
              /* %11110000 */ 0xf0,  75,      
              /* %00001111 */ 0x0f,  75,
              /* %00110011 */ 0x33,  75,
              /* %00001111 */ 0x0f,  75,
              /* %00110011 */ 0x33,  75
};

uint8_t Blasted[] = {
              26,  // number of rows of data below
              /* %00000000 */ 0x00,  50, 
              /* %00011000 */ 0x18,  50,
              /* %00111100 */ 0x3c,  50,
              /* %01111110 */ 0x7e,  50,
              /* %11111111 */ 0xff,  50,
              /* %00000000 */ 0x00,  50, 
              /* %00011000 */ 0x18,  50,
              /* %00111100 */ 0x3c,  50,
              /* %01111110 */ 0x7e,  50,
              /* %11111111 */ 0xff,  50,
              /* %00000000 */ 0x00,  50, 
              /* %00011000 */ 0x18,  50,
              /* %00111100 */ 0x3c,  50,
              /* %01111110 */ 0x7e,  50,
              /* %11111111 */ 0xff,  50,
              /* %00000000 */ 0x00,  50,    
              /* %11111111 */ 0xff,  50,    
              /* %00000000 */ 0x00,  50,
              /* %11111111 */ 0xff,  50,
              /* %00000000 */ 0x00,  50,    
              /* %11111111 */ 0xff,  50,    
              /* %00000000 */ 0x00,  50,
              /* %11111111 */ 0xff,  50,
              /* %00000000 */ 0x00,  50,    
              /* %11111111 */ 0xff,  50,    
              /* %00000000 */ 0x00,  50
};

char RayNelson[]  = "IAIHG TPJNU QU CZR GALWXK DC MHR LANK FOTLA OTN LOYOC HPMPB PX HKICW";
char Test4[]      = "DID YOU REALLY THINK THAT IT WOULD BE SO EASY? Really?  Just running strings?";
char Greets[]     = {16,77,85,66,83,69,67,85,32,74,69,32,84,85,86,83,69,68,32,74,77,85,68,74,79,32,74,77,69,13,0};
char Detective[]  = {13,74,85,82,69,82,32,71,66,32,79,82,84,86,65,32,86,32,88,65,66,74,32,83,86,65,81,32,85,78,69,66,89,81,13,0};
char Scientist[]  = {76,81,84,89,86,70,32,82,75,66,32,83,78,90,32,83,81,87,83,85,32,87,82,65,32,73,77,82,66,32,67,70,72,82,32,90,65,65,65,65,32,73,89,77,87,90,32,80,32,69,65,74,81,86,68,32,89,79,84,80,32,76,71,65,87,32,89,75,90,76,13,0};
char Diver[]      = {10,'D','B','I',' ','D','R','O',' ','P','S','B','C','D',' ','R','K','V','P',' ','Y','P',' ','R','S','C',' ','Z','R','Y','X','O',' ','X','E','W','L','O','B',' ','P','Y','V','V','Y','G','O','N',' ','L','I',' ','R','S','C',' ','V','K','C','D',' ','X','K','W','O',' ','D','R','O','X',' ','D','R','O',' ','C','O','M','Y','X','N',' ','R','K','V','P',' ','Y','P',' ','R','S','C',' ','X','E','W','L','O','B',CR,0};
char Driver[]     = "SOMETIMES WE HAVE ANSWERS AND DONT EVEN KNOW IT SO ENJOY THE VIEW JUST BE HAPPY";
char Politician[] = {83,83,80,87,76,77,32,84,72,67,65,80,32,81,80,32,74,84,32,73,87,69,32,87,68,88,70,90,32,89,85,90,88,32,85,77,86,72,88,72,32,90,65,32,67,66,32,80,65,69,32,88,82,79,76,32,70,65,89,32,73,80,89,75,13,0};
char Test3[]      = {'Z','G','J','G',' ','M','T','M',' ','L','L','P','N',' ','C',' ','N','T','E','R',' ','M','P','M','H',' ','T','W',CR,0};
char Football[]   = "IT MIGHT BE HELPFUL LATER IF YOU KNOW HOW TO GET TO EDEN OR AT LEAST THE WAY";
char Mystery[]    = "OH A MYSTERY STRING I SHOULD HANG ON TO THIS FOR LATER I WONDER WHAT ITS FOR OR WHAT IT DECODES TO?";

char Cmd00[] = { 0x05, 0x42, 0x54, 0x57, 0x50, 0x20, 0x4A, 0x4E, 0x4C, 0x4D, 0x59, 0x20, 0x4D, 0x54,
                 0x5A, 0x57, 0x58, 0x0D, 0x00, 0x4C, 0x4F, 0x56, 0x45, 0x00                         };
char Cmd01[] = { 0x04, 0x41, 0x45, 0x58, 0x47, 0x4C, 0x20, 0x58, 0x5A, 0x0D, 0x00                   };
char Cmd02[] = { 0x0E, 0x47, 0x49, 0x50, 0x41, 0x57, 0x48, 0x0D, 0x00, 0x4C, 0x49, 0x46, 0x45, 0x00 };
char Cmd03[] = { 0x0C, 0x45, 0x46, 0x4D, 0x4B, 0x20, 0x4D, 0x45, 0x58, 0x51, 0x51, 0x42, 0x0D, 0x00 };
char Cmd04[] = { 0x04, 0x53, 0x46, 0x49, 0x43, 0x0D, 0x00, 0x47, 0x69, 0x47, 0x21, 0x00             };
char Cmd05[] = { 0x14, 0x48, 0x49, 0x20, 0x43, 0x48, 0x58, 0x59, 0x4A, 0x59, 0x48, 0x58, 0x59, 0x48,
                 0x4E, 0x20, 0x4E, 0x42, 0x49, 0x4F, 0x41, 0x42, 0x4E, 0x0D, 0x00                   };
char Cmd06[] = { 0x02, 0x50, 0x51, 0x20, 0x4B, 0x4F, 0x43, 0x49, 0x4B, 0x50, 0x43, 0x56, 0x4B, 0x51,
                 0x50, 0x0D, 0x00, 0x4A, 0x6F, 0x6E, 0x6E, 0x79, 0x4D, 0x61, 0x63, 0x00             };
char Cmd07[] = { 0x0C, 0x59, 0x4D, 0x44, 0x44, 0x4B, 0x20, 0x4D, 0x5A, 0x50, 0x20, 0x44, 0x51, 0x42,
                 0x44, 0x41, 0x50, 0x47, 0x4F, 0x51, 0x0D, 0x00, 0x48, 0x41, 0x50, 0x50, 0x59, 0x00 };
char Cmd08[] = { 0x05, 0x4A, 0x46, 0x59, 0x0D, 0x00, 0x48, 0x45, 0x41, 0x4C, 0x54, 0x48, 0x00       };
char Cmd09[] = { 0x09, 0x4D, 0x58, 0x20, 0x57, 0x58, 0x43, 0x20, 0x5A, 0x44, 0x4E, 0x42, 0x43, 0x52,
                 0x58, 0x57, 0x20, 0x4A, 0x44, 0x43, 0x51, 0x58, 0x41, 0x52, 0x43, 0x48, 0x0D, 0x00 };
char Cmd10[] = { 0x0F, 0x52, 0x44, 0x43, 0x48, 0x4A, 0x42, 0x54, 0x0D, 0x00                         };
char Cmd11[] = { 0x02, 0x45, 0x51, 0x50, 0x48, 0x51, 0x54, 0x4F, 0x0D, 0x00                         };
char Cmd12[] = { 0x19, 0x41, 0x54, 0x58, 0x0D, 0x00, 0x57, 0x45, 0x41, 0x4C, 0x54, 0x48, 0x00,
                 0x31, 0x6F, 0x35, 0x37, 0x00                                                       };
 
char *Commands[] = {
                 Cmd00, Cmd01, Cmd02, Cmd03, Cmd04, Cmd05,
                 Cmd06, Cmd07, Cmd08, Cmd09, Cmd10, Cmd11,
                 Cmd12
};

int main()
{
  int idx, last, button, cnt;
  
  setup();                                                      // setup badge io and objects

//  test_ir();
  
  term.tx(CLS);                                                 // clear the terminal
  
  if (BADGE_TYPE >= GOON)                                       // check badge type
    alien_badge();

  // non-Goon badge code here

  while (read_pads() == 0x0) {                                  // wait for a pad press
    idx = (prng.random() >> 1) % 13;
    for (cnt = idx << 1; cnt > 0; --cnt)
      term.tx(' ');
    idx = (prng.random() >> 1) % 13;  
    term.caesar(Commands[idx]);
    if (check_ir())
      break;
    pause(250);
  }

  term.tx(CLS); 
  
  term.caesar(Greets);
  term.tx(CR);
  
  term.otp(Test3, Test4);
  term.tx(CR);
  
  last = -1;                                                    // any button to start
  
  for (;;) {
    do {
      check_ir();                                               // check ir process
      button = read_pads();                                     // wait for input
    } while ((button == 0x0) || (button == last));              // must be new
    last = button;                                              // save for next check
    
    switch (button) {
      case 0x1:
        animation.start(Cylon, 0);                              // start animation
        term.caesar(Detective);                                 // display crypto string
        pause(250);                                             // allow clean button release
        break;
     
      case 0x5:
        animation.start(Chaser, 0);
        term.otp(Scientist, Driver);
        pause(250);
        break;
      
      case 0x7:   
        animation.start(Police, 0);
        term.caesar(Diver);
        pause(250);
        break;
        
      case 0x8:
        animation.start(InOut, 0);
        term.caesar(Detective);                                 // display crypto string
        term.otp(Politician, Football);
        pause(250);
        break;
      
      case 0x9:
        animation.stop();
        term.otp(RayNelson, Mystery);
        pause(250);
        break;
      }
  }
    
  return 0;
}

void test_ir()
{
  static int IR_Tests[] = {
    0xA5A,
    0x5A5,
    0x00F,
    0x0F0,
    0xF00
  };
  
  int checks = 0;
  int idx, cnt;
  
  term << "Starting IR test...\r";
  irtx.start(IR_OUT, IR_FREQ);
  irrx.start(IR_IN);               
  irrx.enable();
  term <<"IR initialization done\r";

  for (idx = 0; idx <= 4; ++idx) {
    for (cnt = 3; cnt > 0; --cnt) {
      irrx.enable();
      irtx.tx(IR_Tests[idx], 12, 1);
      while (irtx.busy())                                       // hold for frame transmission
        ;
      if (irrx.bit_count() == 12) {                             // good bit count?
        if (irrx.rxcheck() == IR_Tests[idx]) {                  // good code?
          checks += 1;
          break;
        }
      }
    }
  }
          
  if (checks == 5)
    term << "IR test passed!\r";
  else
    term << "IR test failed\r";
    
  for (;;)
    ;
}

int check_ir()
{
  int code;

  code = irrx.rxcheck();                                        // check ir input
  
  if (code < 0)                                                 // abort if waiting
    return false;

  if (code == IR_BLAST) {
    animation.start(Blasted, 3);                                                      
    term.caesar("\nDROI VSFO  GO CVOOZ\r");                     // THEY LIVE  WE SLEEP  
    while (animation.running())                                 // let animation finish
      ;
  }  
            
  irrx.enable();

  return (code == IR_BLAST);
}

void alien_badge()
{
  int delay;

// Badge code for >= Goon

  delay = -1;                                                   // force immediate broadcast

  while (1) {
    read_pads(); // dbetz -- not sure what this is for
    if ((read_pads()) || (tmr1.seconds() >= delay)) {           // pad input or delay expired?
      irtx.tx(IR_BLAST, 16, 3);                                 // blast the humans!    
      term.caesar("\x0cIQ XUHQ  FTQK EXQQB\r");                 // WE LIVE  THEY SLEEP                                          
      animation.start(Blasted, 3);                              // play blaster animation    
      tmr1.start();                                             // restart timer               
      delay = ABS(prng.random()) % 11 + 10;                     // new delay
    }
  }
}

void setup()
{
// Setup badge IO and objects
// -- set speed before starting other objects

  set_speed(B_SPEED);                                           // set badge speed (MHz)
  
  term.start(RX1, TX1, CryptoFdSerial::MODE_NONE, T_BAUD);      // start terminal

  prng.seed(CNT << 2, CNT, 0x1057, -CNT, (int32_t)CNT >> 2);    // seed prng (random #s)

  if (BADGE_TYPE < GOON) {
    irrx.start(IR_IN);                                          // start ir input               
    irrx.enable();                                              // accept ir input now
  }        
  else {
    irtx.start(IR_OUT, IR_FREQ);                                // start ir output  
    tmr1.start();                                               // start timer
  }
}

/* ----------------------------- */
/*  B A D G E   F E A T U R E S  */
/* ----------------------------- */

void set_speed(int mhz)
{
// Sets badge clock speed
// -- sets timing variables ms001 and us001
// -- note: objects may require restart after speed change

  switch (mhz) {
    case  0: clkset(RC_SLOW,    20000); break;                  // super low power -- sleep mode only!
    case  5: clkset(XT1_PL1,  5000000); break;
    case 10: clkset(XT1_PL2, 10000000); break;
    case 20: clkset(XT1_PL4, 20000000); break;
    case 40: clkset(XT1_PL8, 40000000); break; 
    case 80: clkset(XT1_P16, 80000000); break;
  }

  waitcnt(CNT + (CLKFREQ / 100));                               // wait ~10ms

  ms001 = CLKFREQ / 1000;                                       // set ticks per millisecond for waitcnt
  us001 = CLKFREQ / 1000000;                                    // set ticks per microsecond for waitcnt
}

//                 0x0, 0x1, 0x2, 0x3, 0x4, 0x5, 0x6, 0x7, 0x8, 0x9, 0xa, 0xb, 0xc, 0xd, 0xe, 0xf
uint8_t rev4[] = { 0x0, 0x8, 0x4, 0xc, 0x2, 0xa, 0x6, 0xe, 0x1, 0x9, 0x5, 0xd, 0x3, 0xb, 0x7, 0xf };

int read_pads()
{
  uint32_t mask = 0xf << PAD0;

// Reads and returns state of touch pad inputs
// -- swaps LSB/MSB for correct binary input

  OUTA |= mask;                                                 // charge pads (all output high)   
  DIRA |= mask;
    
  DIRA &= ~mask;                                                // float pads   
  pause(50);                                                    // -- allow touch to discharge
 
#ifdef QUICKSTART
  return ((~INA & mask) >> PAD0) & 0xf;                         // return "1" for touched pads
#else
  return rev4[((~INA & mask) >> PAD0) & 0xf];                   // return "1" for touched pads
#endif
}

/* ------------- */
/*  B A S I C S  */
/* ------------- */

void pause(int ms)
{
  int t;

// Delay program in milliseconds
// -- ensure set_speed() used before calling

  t = CNT;                                                      // sync to system counter
  while (--ms >= 0)                                             // delay > 0
    waitcnt(t += ms001);                                        // hold 1ms
}

void high(int pin)
{
  uint32_t mask = 1 << pin;
  
// Makes pin output and high

  OUTA |= mask;
  DIRA |= mask;
}

void low(int pin)
{
  uint32_t mask = 1 << pin;
  
// Makes pin output and low

  OUTA &= ~mask;
  DIRA += mask;
}

void toggle(int pin)
{
  uint32_t mask = 1 << pin;

// Toggles pin state

  OUTA ^= mask;
  DIRA |= mask;
}

int input(int pin)
{
  uint32_t mask = 1 << pin;

// Makes pin input and returns current state

  DIRA &= ~mask;

  return (INA & mask) != 0;
}

#if 0

pub pulse_out(pin, us) | state

// Generate pulse on pin for us microseconds
// -- ensure set_speed() used before calling
// -- makes pin output
// -- pulse out is opposite of pin//s input state
// -- blocks until pulse is finished (to clear counter)

  us *= us001                                                   // convert us to system ticks
  state := ina[pin]                                             // read incoming state of pin

  if (ctra == 0)                                                // ctra available?
    if (state == 0)                                             // low-high-low
      low(pin)                                                  // set to output
      frqa := 1 
      phsa := -us                                               // set timing
      ctra := (%00100 << 26) | pin                              // start the pulse
      repeat
      until (phsa => 0)                                         // let pulse finish
                          
    else                                                        // high-low-high
      high(pin)
      frqa := -1
      phsa := us
      ctra := (%00100 << 26) | pin
      repeat
      until (phsa < 0) 

    ctra := IS_OFF                                              // release counter
    return true

  elseif (ctrb == 0)               
    if (state == 0)                
      low(pin)                     
      frqb := 1 
      phsb := -us               
      ctrb := (%00100 << 26) | pin 
      repeat
      until (phsb => 0)            
                          
    else                           
      high(pin)
      frqb := -1
      phsb := us
      ctrb := (%00100 << 26) | pin
      repeat
      until (phsb < 0) 

    ctrb := IS_OFF                     
    return true
  
  else
    return false                                                // alert user of error
  

pub set_freq(ctrx, px, fx)

// Sets ctrx to frequency fx on pin px (NCO/SE mode)
// -- fx in hz
// -- use fx of 0 to stop counter that is running

  if (fx > 0)                             
    fx := ($8000_0000 / (CLKFREQ / fx)) << 1                    // convert freq for NCO mode    
    case ctrx                                                    
      "a", "A":                                                  
        ctra := ((%00100) << 26) | px                           // configure ctra for NCO on pin
        frqa := fx                                              // set frequency
        dira[px] := IS_OUTPUT                                    
                                                                 
      "b", "B":                                                  
        ctrb := ((%00100) << 26) | px                            
        frqb := fx                                               
        dira[px] := IS_OUTPUT                                    
                                                                 
  else                                                           
    case ctrx                                                    
      "a", "A":                                                  
        ctra := IS_OFF                                          // disable counter
        outa[px] := IS_OFF                                      // clear pin/driver 
        dira[px] := IS_INPUT                                  
     
      "b", "B":                         
        ctrb := IS_OFF  
        outa[px] := IS_OFF  
        dira[px] := IS_INPUT 

#endif


/*

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/